class MockupView : Gtk.DrawingArea
{
    private Mockup mockup;
    private double select_x;
    private double select_y;
    private bool resizing = false;
    private bool dragging = false;

    public MockupView (Mockup mockup)
    {
        this.mockup = mockup;
        add_events (Gdk.EventMask.BUTTON_PRESS_MASK | Gdk.EventMask.BUTTON1_MOTION_MASK | Gdk.EventMask.BUTTON_RELEASE_MASK);
        mockup.item_added.connect (item_added_cb);
        foreach (var item in mockup.items)
            item_added_cb (item);
    }

    public override void get_preferred_width (out int minimum_width, out int natural_width)
    {
        minimum_width = natural_width = 300;
    }

    public override void get_preferred_height (out int minimum_height, out int natural_height)
    {
        minimum_height = natural_height = 200;
    }

    public override bool draw (Cairo.Context c)
    {
        c.set_source (get_background_pattern (c));
        c.paint ();

        foreach (var item in mockup.items)
            item.draw (c);

        /* Draw selection box */
        var item = mockup.selected_item;
        if (item != null)
        {
            c.rectangle (item.x + 0.5, item.y + 0.5, item.width, item.height);
            c.set_source_rgb (0.0, 0.0, 0.0);
            c.set_line_width (1.0);
            c.set_dash ({2.0, 3.0}, 0.5);
            c.stroke ();
            c.set_dash (null, 0.0);

            if (item.resizable)
            {
                c.rectangle (item.x + item.width - 3.5, item.y + item.height - 3.5, 7, 7);
                c.set_source_rgb (1.0, 1.0, 1.0);
                c.fill_preserve ();
                c.set_line_width (1.0);
                c.set_source_rgb (0.0, 0.0, 0.0);
                c.stroke ();
            }

            foreach (var p in item.control_points)
            {
                c.arc (p.x, p.y, p.radius, 0, 2 * Math.PI);
                c.set_source_rgb (1.0, 1.0, 1.0);
                c.fill_preserve ();
                c.set_line_width (1.0);
                c.set_source_rgb (0.0, 0.0, 0.0);
                c.stroke ();
            }
        }

        return true;
    }

    public override bool button_press_event (Gdk.EventButton event)
    {
        var x = (int) event.x;
        var y = (int) event.y;

        /* If a selected item see if resizing */
        // FIXME: Replace with control point
        if (mockup.selected_item != null && mockup.selected_item.resizable)
        {
            var item = mockup.selected_item;
            var handle_x = item.x + item.width - 3;
            var handle_y = item.y + item.height - 3;
            var dx = x - handle_x;
            var dy = y - handle_y;
            if (dx >= 0 && dx <= 7 && dy >= 0 && dy <= 7)
            {
                dragging = true;
                resizing = true;
                select_x = event.x;
                select_y = event.y;
                return true;
            }
        }
        resizing = false;

        var items = mockup.items.copy ();
        items.reverse ();

        foreach (var item in items)
        {
            foreach (var p in item.control_points)
            {
                var dx = x - p.x;
                var dy = y - p.y;
                if (dx*dx + dy*dy < p.radius*p.radius)
                {
                    mockup.select_item (item, p);
                    select_x = event.x;
                    select_y = event.y;
                    dragging = true;
                    queue_draw ();
                    return true;
                }
            }
        }

        foreach (var item in items)
        {
            if (item.select (x, y))
            {
                mockup.select_item (item);
                select_x = event.x;
                select_y = event.y;
                dragging = true;
                queue_draw ();
                return true;
            }
        }

        mockup.select_item (null);
        queue_draw ();
        return false;
    }
    
    public override bool motion_notify_event (Gdk.EventMotion event)
    {
        if (!dragging || mockup.selected_item == null)
            return false;

        var dx = event.x - select_x;
        var dy = event.y - select_y;

        var x_step = (int) Math.trunc (dx / 10.0);
        var y_step = (int) Math.trunc (dy / 10.0);
        if (x_step == 0 && y_step == 0)
            return false;

        if (resizing)
            mockup.resize (x_step * 10, y_step * 10);
        else
            mockup.move (x_step * 10, y_step * 10);
        select_x += x_step * 10;
        select_y += y_step * 10;
        queue_draw ();

        return true;
    }

    public override bool button_release_event (Gdk.EventButton event)
    {
        if (!dragging)
            return false;

        dragging = false;
        return true;
    }

    private void item_added_cb (MockItem item)
    {
        item.redraw.connect (() => { queue_draw (); });
        queue_draw ();
    }

    private Cairo.Pattern? _background_pattern = null;
    private Cairo.Pattern get_background_pattern (Cairo.Context c)
    {
        if (_background_pattern != null)
            return _background_pattern;

        var surface = new Cairo.Surface.similar (c.get_target (), Cairo.Content.COLOR, 10, 10);
        var grid_c = new Cairo.Context (surface);
        grid_c.set_source_rgb (1.0, 1.0, 1.0);
        grid_c.paint ();

        grid_c.set_source_rgb (0.0, 0.0, 0.0);
        grid_c.rectangle (0.0, 0.0, 1.0, 1.0);
        grid_c.fill ();

        _background_pattern = new Cairo.Pattern.for_surface (surface);
        _background_pattern.set_extend (Cairo.Extend.REPEAT);

        return _background_pattern;
    }
}
