public class Mockup : Object
{
    // FIXME: Replace with a tree that automatically reparents when dragging
    public List<MockItem> items;
    public MockItem? selected_item = null;
    public MockItemControlPoint? selected_control_point = null;
    public List<MockItem> selected_child_items;

    public signal void item_added (MockItem item);
    public signal void selected_item_changed ();

    public Mockup ()
    {
        items = new List<MockItem> ();
    }

    public void add_item (MockItem item)
    {
        insert (item);
        item_added (item);

        if (selected_item != null && item.depth > selected_item.depth)
        {
            item.x = selected_item.x;
            item.y = selected_item.y;
        }

        select_item (item);
    }

    private void insert (MockItem item)
    {
        unowned List<MockItem> iter;
        for (iter = items; iter != null; iter = iter.next)
        {
            if (iter.data.depth > item.depth)
                break;
        }
        items.insert_before (iter, item);
    }

    public void select_item (MockItem? item, MockItemControlPoint? control_point = null)
    {
        if (selected_item == item && selected_control_point == control_point)
            return;
        selected_item = item;
        selected_control_point = control_point;

        selected_child_items = new List<MockItem> ();
        if (item != null)
        {
            /* Work out which items are inside this one */
            foreach (var i in items)
            {
                if (i == item)
                    continue;

                if (item.contains (i))
                    selected_child_items.append (i);
            }

            /* Raise to the top of this item class */
            items.remove (item);
            insert (item);
        }

        selected_item_changed ();
    }

    public void move (int dx, int dy)
    {
        if (selected_item == null)
            return;
            
        if (selected_control_point != null)
        {
            selected_control_point.x += dx;
            selected_control_point.y += dy;
            selected_control_point.moved ();
            return;
        }

        /* Move this item and its children */
        selected_item.x += dx;
        selected_item.y += dy;
        foreach (var item in selected_child_items)
        {
            item.x += dx;
            item.y += dy;
        }
    }

    public void resize (int dx, int dy)
    {
        if (selected_item == null)
            return;

        /* Move this item and its children */
        selected_item.width += dx;
        if (selected_item.width < 10)
            selected_item.width = 10;
        selected_item.height += dy;
        if (selected_item.height < 10)
            selected_item.height = 10;
    }
}

public abstract class MockItem
{
    public int x;
    public int y;
    public int width;
    public int height;
    public bool resizable = false;
    public int depth;
    public List<MockItemProperty> properties;
    public List<MockItemControlPoint> control_points;

    public signal void redraw ();

    public abstract void draw (Cairo.Context c);

    public virtual bool select (int x, int y)
    {
        var dx = x - this.x;
        var dy = y - this.y;
        return dx >= 0 && dy >= 0 && dx < width && dy < height;
    }

    protected void add_property (MockItemProperty property)
    {
        properties.append (property);
    }

    protected void add_control_point (MockItemControlPoint control_point)
    {
        control_points.append (control_point);
    }

    public bool contains (MockItem item)
    {
        /* Only allow higher things into here */
        if (item.depth <= depth)
            return false;

        return item.x >= x && (item.x + item.width) <= (x + width) && item.y >= y && (item.y + item.height) <= (y + height);
    }
}

public abstract class MockItemProperty
{
    public string id;
    public string description;

    public signal void value_changed ();
    
    public MockItemProperty (string id, string description)
    {
        this.id = id;
        this.description = description;
    }

    public abstract string to_string ();
    public abstract void set_from_string (string text);
}

public class MockItemStringProperty : MockItemProperty
{
    public string _value = "";
    public string value
    {
        get { return _value; }
        set
        {
            _value = value;
            value_changed ();
        }
    }

    public MockItemStringProperty (string id, string description, string value)
    {
        base (id, description);
        _value = value;
    }

    public override string to_string ()
    {
        return value;
    }

    public override void set_from_string (string text)
    {
        value = text;
    }
}

public class MockItemBooleanProperty : MockItemProperty
{
    public bool _value = false;
    public bool value
    {
        get { return _value; }
        set
        {
            _value = value;
            value_changed ();
        }
    }
    
    public MockItemBooleanProperty (string id, string description, bool value)
    {
        base (id, description);
        _value = value;
    }

    public override string to_string ()
    {
        return value ? "true" : "false";
    }

    public override void set_from_string (string text)
    {
        value = text == "true";
    }
}

public class MockItemIntegerProperty : MockItemProperty
{
    public int _value = 0;
    public int value
    {
        get { return _value; }
        set
        {
            if (value < min)
                value = min;
            else if (value > max)
                value = max;
            _value = value;
            value_changed ();
        }
    }
    public int min;
    public int max;
    
    public MockItemIntegerProperty (string id, string description, int value, int min, int max)
    {
        base (id, description);
        _value = value;
        this.min = min;
        this.max = max;
    }

    public override string to_string ()
    {
        return value.to_string ();
    }

    public override void set_from_string (string text)
    {
        value = int.parse (text);
    }
}

public class MockItemControlPoint
{
    public int x;
    public int y;
    public double radius;

    public signal void moved ();

    public MockItemControlPoint (int x, int y, double radius = 3.5)
    {
        this.x = x;
        this.y = y;
        this.radius = radius;
    }
}
